@@ -22,6 +22,8 @@ EDWIN edwin[4];
 int cur_y,cur_x,win_x,win_y;            /* (x,y) on screen */
 int cursor_row=0,cursor_col=0;  /* in data */
 int insert_mode=1,insert_mode_=-1;
+int DispTab=1;
+int TabChar=1;
 
 int page_row=0,page_col=0;
 int page_len=M_ROW-2,page_width=M_COL;
@@ -33,6 +35,7 @@ void init_edbf()
 {
 if ((edbf=mmalloc(sizeof(LIN)*INI_Alloc_Lines,"init_edbf"))==NULL)
 	exit(-1);
+edstate[cur_file].edbf=edbf;
 Alloc_Lines=INI_Alloc_Lines;
 }
 
@@ -42,8 +45,8 @@ Lines++;
 if (Alloc_Lines<=Lines+1) {
 	Alloc_Lines+=200;
 	if ((edbf=(LIN *)mrealloc(edbf,
-		sizeof(LIN)*Alloc_Lines,"incr_line"))==NULL)
-	exit(-1);
+		sizeof(LIN)*Alloc_Lines,"incr_line"))==NULL)	exit(-1);
+	edstate[cur_file].edbf=edbf;
 }
 }
 
@@ -51,9 +54,10 @@ void decr_line()
 {
 Lines--;
 if (Alloc_Lines>Lines+201) {
-Alloc_Lines-=200;
-if ((edbf=(LIN *)mrealloc(edbf,sizeof(LIN)*Alloc_Lines,"err:decr_line"))==NULL)
-	exit(-1);
+	Alloc_Lines-=200;
+	if ((edbf=(LIN *)mrealloc(edbf,sizeof(LIN)*Alloc_Lines,"err:decr_line"))
+		==NULL) exit(-1);
+	edstate[cur_file].edbf=edbf;
 }
 }
 
@@ -62,9 +66,10 @@ void ins_lines(int row, int lines)
 int i;
 Lines+=lines;
 if (Alloc_Lines<=Lines+1) {
-Alloc_Lines+=Max(lines,200);
-if ((edbf=(LIN *)mrealloc(edbf,sizeof(LIN)*Alloc_Lines,"err:ins_lines"))==NULL)
-	exit(-1);
+	Alloc_Lines+=Max(lines,200);
+	if ((edbf=(LIN *)mrealloc(edbf,
+		 sizeof(LIN)*Alloc_Lines,"err:ins_lines"))==NULL) exit(-1);
+	edstate[cur_file].edbf=edbf;
 }
 for(i=Lines-lines-1;i>=row;i--) edbf[i+lines]=edbf[i];
 }
@@ -113,7 +118,7 @@ int i,j,k;
 char *p,ttt[20];
 if (!len) return;
 for(p=len+fname-1; p>=fname && *p!='.'; p--) ;
-if (*p!='.') {
+if (p<fname) {
 	i=LangN;
 	goto search_1st_line;
 }
@@ -166,7 +171,7 @@ next_1:
 	}
 	goto nomatch;
 }
-if (strstr(".c .h .cc .java",p)) c_comment=1;
+if (strstr(".c .h .cc .cpp .java",p)) c_comment=1;
 else c_comment=0;
 match_lang:
 LangType=i;
@@ -239,7 +244,11 @@ for(;wc-page_col<page_width && i<len;i++) {
 	 case 9:
 	   dd=tab_width-(wc%tab_width);
 	   wc+=dd;
-	   for(j=0;j<dd;j++)xputch_w(' ');
+	   if (DispTab)  {
+			 xputch_w(TabChar);
+	     for(j=1;j<dd;j++)xputch_w(' ');
+	   } else
+	     for(j=0;j<dd;j++)xputch_w(' ');
 	   break;
 	 default:
 	   xputch(lin[i]);
@@ -317,24 +326,7 @@ if (cmt_line) {
 } else
 if (len && lin[0]==DirectCommentChar) tatt=kwp[0].coloridx;
 
-for(wc=i=0;i<len;) {
-	if (wc >= page_col) break;
-	if (lin[i]==9)
-	 wc+=tab_width-(wc%tab_width);
-	else {
-		wc++;
-		if (lin[i]=='"' || lin[i]=='\'') {
-			if (!inquote) inquote=1;
-			else
-			if (i&& lin[i-1]!='\\') inquote=0;
-		}
-#if	0
-		else
-		if (lin[i]==DirectCommentChar) tatt=keywcolor[0];
-#endif
-	}
-	i++;
-}
+wc=0;
 #if	1
 if (in_mblock(row,wc)) {
 	set_att(3);
@@ -344,10 +336,10 @@ if (in_mblock(row,wc)) {
 if (inquote) set_att(kwp[1].coloridx);
 else
 if (!incmt) set_att(tatt);
+
 if (inlmark) set_att(3);
-for(j=0;j<wc-page_col;j++)
-	xputch(' ');
-for(;wc-page_col<page_width && i<len;) {
+
+for(wc=i=0;wc-page_col<page_width && i<len;) {
 	if (in_mblock(row,i)) {
 		inmark=1;
 		set_att(3);
@@ -360,16 +352,26 @@ for(;wc-page_col<page_width && i<len;) {
 	switch (lin[i]) {
 	 case 9:
 	   dd=tab_width-(wc%tab_width);
-	   wc+=dd;
 	   if (inlmark||inmark) set_att(3);
-	   for(j=0;j<dd;j++)xputch_w(' ');
-	   break;
+	   if (DispTab)  {
+	     if (wc++>=page_col)
+			   xputch_w(TabChar);
+	     for(j=1;j<dd;j++)
+	       if (wc++>=page_col)
+	         xputch_w(' ');
+	   } else
+	     for(j=0;j<dd;j++)
+	       if (wc++>=page_col)
+	         xputch_w(' ');
+	   i++;
+	   continue;
 	 default:
 	if (!inquote && c_comment && i<len-1
 		&& (lin[i]=='/' && (lin[i+1]=='*' || lin[i+1]=='/') ||
 		lin[i]=='*' && lin[i+1]=='/') ) {
 		set_att(kwp[2].coloridx);
-		mxputch(lin[i]);
+		if (wc++>=page_col)
+		  mxputch(lin[i]);
 		if (lin[i]=='/' && lin[i+1]=='/') {
 			incmt=2;
 		} else
@@ -378,41 +380,41 @@ for(;wc-page_col<page_width && i<len;) {
 	else
 	if (!inquote && c_comment && incmt==1 && i &&
 				lin[i-1]=='*' && lin[i]=='/') {
-		mxputch(lin[i]);
+		if (wc++>=page_col)
+		  mxputch(lin[i]);
 		incmt=0;
 		set_att(0);
 	}
 	else
 	if (incmt) {
-		mxputch(lin[i]);
-		goto lll2;
+		if (wc++>=page_col)
+		  mxputch(lin[i]);
 	} else
 	/* for the # char in perl and shell */
-#if	0
-	if (!c_comment && lin[i]==DirectCommentChar && !inquote) {
-#else
 	if (!c_comment && lin[0]==DirectCommentChar && !inquote) {
-#endif
 		set_att(kwp[0].coloridx);
-		mxputch(lin[i]);
+		if (wc++>=page_col)
+		  mxputch(lin[i]);
 		incmt=2;
-		goto lll2;
 	} else
 	if (lin[i]=='"' || lin[i]=='\'') {
 		if (!inquote) {
 			inquote=lin[i];
 			set_att(kwp[1].coloridx);
-			mxputch(lin[i]);
+		  if (wc++>=page_col)
+			  mxputch(lin[i]);
 		}
 		else {
 			if (!basl) {
-				mxputch(lin[i]);
+		    if (wc++>=page_col)
+				  mxputch(lin[i]);
 				if (lin[i]==inquote) {
 					inquote=0;
 					set_att(0);
 				}
 			} else {
-				mxputch(lin[i]);
+		    if (wc++>=page_col)
+				  mxputch(lin[i]);
 				basl=0;
 			}
 		}
@@ -453,8 +455,9 @@ lll6:
 					inmark=0;
 					set_att(tatt);
 				}
-				mxputch(*p);
-				wc++;
+		    if (wc++>=page_col)
+				  mxputch(*p);
+				/* wc++; */
 				i++;
 			}
 			set_att(0);
@@ -464,10 +467,9 @@ lll6:
 			set_att(0);
 		}
 lll1:
-		mxputch(lin[i]);
+		if (wc++>=page_col)
+		  mxputch(lin[i]);
 	}
-lll2:
-		wc++;
 	}
 	i++;
 lll5:
@@ -802,11 +804,13 @@ void f_modified_quit_getch(int ch)
 		close_file(0);
 }
 
+int FModifiedQuitYN=1;
 void close_quit_file()
 {
 	if (f_modified) {
 		message(4,"Not saved. Quit (Y/N):",fname);
-		setup_getch(23,MROW-1,"YNyn",'Y',f_modified_quit_getch);
+		setup_getch(23,MROW-1,"YNyn",FModifiedQuitYN?'Y':'N',
+		  f_modified_quit_getch);
 	}
 	else
 	close_file();
@@ -916,6 +920,7 @@ if (dire) {
 	cursor_col=i;
 	pos_cur();
 	disp_page();
+	redraw_status();
 	show_cursor_w();
 	return;
 } else
@@ -927,6 +932,7 @@ if (!dire) {
 	cursor_col=i;
 	pos_cur();
 	disp_page();
+	redraw_status();
 	show_cursor_w();
 	return;
 }
@@ -1315,6 +1321,13 @@ K_FileBrowser()
 	return 1;
 }
 
+K_TogDispTab()
+{
+DispTab^=1;
+disp_page();
+show_cursor_w();
+}
+
 void handle_key(char *astr, int skey, int state)
 {
 	if (input_handler==1) {
@@ -1360,13 +1373,13 @@ void setup_xgets()
 	write_blk_init();
 }
 
-static void ign_intr()
+static void ign_intr(int a)
 {
 signal(SIGINT,ign_intr);
 }
 
 int fnamesN=0, fnamesN_a=10;
-char *fnames[256];
+char *fnames[512];
 int fgps=0,bgps=1;
 
 void put_arg_fname(char *s)
@@ -1374,6 +1387,8 @@ void put_arg_fname(char *s)
 fnames[fnamesN++]=strdup(s);
 }
 
+char *exec_name;
+
 main(int argc, char **argv)
 {
 int i;
@@ -1381,6 +1396,10 @@ extern int cdtedrc;
 extern int ReLoadTed;
 extern char *ini_fcmd,*goto_tagname;
 
+set_kdef();
+exec_name=argv[0];
+K_LoadTedrc(tedrcfname);
+
 init_X_arg(argc,argv);
 #ifndef DEBUG
 if (!fgps) fgps=!bgps;
@@ -1390,8 +1409,6 @@ if (!fgps) {
 	if (pid>0) exit(0);
 }
 #endif
-set_kdef();
-K_LoadTedrc(tedrcfname);
 if (!batch) init_X_win(argc,argv);
 get_cdir();
 setup_xgets();
@@ -1404,12 +1421,11 @@ if (!fnamesN) {
 		if (lookup_tag(goto_tagname))
 			p_err("cannot lookup tag '%s'",goto_tagname);
 	} else
-	edit_file();
+	edit_file(fname);
 }
 else {
 	for (i=0;i<fnamesN;i++) {
-		strcpy(fname,fnames[i]);
-		edit_file();
+		edit_file(fnames[i]);
 	}
 	goto_file(0);
 }
