@@ -15,17 +15,12 @@
 #define SEEK_CUR (2)
 #endif
 
-typedef struct {
-	char fname[128];
-	int page_row,cursor_row,cursor_col;
-	int wx,wy,wrow,wcol;
-} FPOS;
-
 int file_no,nosave;
 extern int MAX_FILE;
 char tedfilepos[]=".tedfilepos";
 
 char fname[256]="#noname";
+char fdir[128];
 
 mfgets(char *s,int len,FILE *fp)
 {
@@ -115,9 +110,10 @@ edstate[f_no].f_modified=f_modified;
 edstate[f_no].nosave=nosave;
 edstate[f_no].readonly=readonly;
 edstate[f_no].edbf=edbf;
-edstate[f_no].Lines=Lines;
 edstate[f_no].Alloc_Lines=Alloc_Lines;
+edstate[f_no].Lines=Lines;
 strcpy(edstate[f_no].fname,fname);
+strcpy(edstate[f_no].fdir,fdir);
 edstate[f_no].cmapno=cmapno;
 }
 
@@ -125,67 +121,78 @@ edstate[f_no].cmapno=cmapno;
 static int fds[2];
 int StripCR=1;
 
-void read_file(int pope)
+read_file(char *fn, int pope)
 {
 char *bf,*str;
-int len, lc;
+int len;
 int pid=0,status;
 int fd;
 
-init_edbf();
-edbf[0].cstr=0;
-edbf[0].len=0;
 if (pope) {
 	if (pipe(fds)<0) {
 		error("Cannot create pipe");
-		return;
+		return -1;
 	}
 	if (!(pid=fork())) {
 		dup2(fds[1],1);
 		dup2(fds[1],2);
 		close(fds[0]);
-		if (strchr(fname,' '))
-			execl("/bin/sh","sh","-c",fname,NULL);
-		else    execlp(fname,fname,NULL);
+		if (strchr(fn,' '))
+			execl("/bin/sh","sh","-c",fn,NULL);
+		else    execlp(fn,fn,NULL);
 	} else
 	if (pid<0) {
 		error("Cannot fork process");
-		return;
+		return -1;
 	}
 	close(fds[1]);
 	fd=fds[0];
 }
 else
-if (!strcmp(fname,"-")) fd=0;
+if (!strcmp(fn,"-")) fd=0;
 else
-if ((fd=open(fname,O_RDONLY))<0)
+if ((fd=open(fn,O_RDONLY))<0)
 		bell();
 
+if (fn[0]!='/') {
+	getcwd(fdir,sizeof(fdir));
+} else {
+	strcpy(fdir,"/");
+}
+
+if (!pope && file_no<20) {
+	if (test_file(fdir,fn)) {
+    if (cur_file>=file_no)
+	    cur_file=file_no-1;
+	  return -1;
+	}
+}
+strcpy(fname,fn);
+
+init_edbf();
+edbf[0].cstr=0;
+edbf[0].len=0;
 if(fd<0) {
 	Lines=1;
 	disp_win_title();
-	return;
+	return 0;
 }
-lc=Lines=0;
+Lines=0;
 init_fdbuf(sizeof(fdbf));
 for(;;) {
 	len=fdgets(&bf,fd);
 	if (!len) {
-		edbf[lc].len=0;
-		edbf[lc].cstr=0;
+		edbf[Lines].len=0;
+		edbf[Lines].cstr=0;
 		break;
 	}
 	if (bf[len-1]=='\n') bf[--len]=0;
 	   	if (StripCR && len && bf[len-1]=='\r') bf[--len]=0;
-	edbf[lc].len=len;
-	if (!len) {
-		edbf[lc].cstr=0;
-		lc++;
-		incr_line();
-		continue;
-	}
-	edbf[lc].cstr=bf;
-	lc++;
+	edbf[Lines].len=len;
+	if (!len)
+		edbf[Lines].cstr=0;
+	else
+	  edbf[Lines].cstr=bf;
 	incr_line();
 }
 close(fd);
@@ -198,6 +205,7 @@ if (Lines==0) {
 nosave=pope;
 disp_win_title();
 K_SwitchKeyDef(0);
+return 0;
 }
 
 int BackupSave;
@@ -277,6 +285,7 @@ edbf=edstate[f_no].edbf;
 Lines=edstate[f_no].Lines;
 Alloc_Lines=edstate[f_no].Alloc_Lines;
 strcpy(fname,edstate[f_no].fname);
+strcpy(fdir,edstate[f_no].fdir);
 set_lang_type(fname);
 K_SwitchKeyDef(edstate[f_no].cmapno);
 }
@@ -327,6 +336,45 @@ if (file_no>=MAX_FILE) {
 return 0;
 }
 
+int getFileWininfo(char *fname, FPOS *fposp)
+{
+FILE *fp;
+int match=0;
+struct stat st;
+static char tedfpos[]=".tedfilepos";
+
+stat(tedfpos,&st);
+if (st.st_size%sizeof(FPOS)) {
+  unlink(tedfpos);
+  return 0;
+}
+if ((fp=fopen(tedfpos,"r"))==NULL)
+  return 0;
+
+bzero(fposp,sizeof(FPOS));
+while (!feof(fp)) {
+  fread(fposp,sizeof(FPOS),1,fp);
+  if (!strcmp(fposp->fname,fname)) {
+    match=1;
+    break;
+  }
+}
+fclose(fp);
+if (match) {
+  endian_conv((char *)&fposp->page_row);
+  endian_conv((char *)&fposp->cursor_row);
+  endian_conv((char *)&fposp->cursor_col);
+  endian_conv((char *)&fposp->wrow);
+  endian_conv((char *)&fposp->wcol);
+  endian_conv((char *)&fposp->wx);
+  endian_conv((char *)&fposp->wy);
+  endian_conv((char *)&fposp->fntidx);
+  return 1;
+}
+
+return 0;
+}
+
 char *goto_tagname=NULL;
 
 int vi_goto_no=-1;
@@ -335,68 +383,64 @@ int RestCursor,RestWinPos,RestWinSize;
 
 void close_file_id();
 
-void edit_file()
+edit_file(char *fn)
 {
 FILE *fp;
-extern int c_comment;
+extern int c_comment, mult_open;
+struct stat sta;
 
-if (extend_edstate()) return;
-
-
-bzero(&edstate[file_no],sizeof(edstate[0]));
-page_row=page_col=cursor_row=cursor_col=cur_y=cur_x=f_modified=0;
-/* insert_mode=1; */
-cur_file=file_no;
-read_file(0);
+if (extend_edstate()) return -1;
 save_win_context(active_win);
+cur_file=file_no;
+/* insert_mode=1; */
+if (read_file(fn,0)) return -1;
 file_no++;
+bzero(&edstate[cur_file],sizeof(edstate[0]));
+readonly=page_row=page_col=cursor_row=cursor_col=cur_y=cur_x=f_modified=0;
+
+if (!stat(fn,&sta)) {
+  if (fp=fopen(fn,"a"))
+    fclose(fp);
+  else
+    readonly=1;
+}
 
 if (vi_goto_no >=0) {
 	goto_line(vi_goto_no);
+	save_edstate();
 	vi_goto_no=-1;
-} else
-if (fp=fopen(".tedfilepos","r")) {
-	FPOS fpos;
-	int match=0;
+} else {
+  FPOS fpos;
+  extern int fontidx;
 
-	bzero(&fpos,sizeof(fpos));
-	while (!feof(fp)) {
-		fread(&fpos,sizeof(FPOS),1,fp);
-		if (!strcmp(fpos.fname,fname)) {
-			match=1;
-			break;
-		}
-	}
-	fclose(fp);
-	if (match) {
-		endian_conv((char *)&fpos.page_row);
-		endian_conv((char *)&fpos.cursor_row);
-		endian_conv((char *)&fpos.cursor_col);
-		endian_conv((char *)&fpos.wrow);
-		endian_conv((char *)&fpos.wcol);
-		endian_conv((char *)&fpos.wx);
-		endian_conv((char *)&fpos.wy);
-		if (RestCursor) {
-			page_row=fpos.page_row;
-			cursor_row=fpos.cursor_row;
-			cursor_col=fpos.cursor_col;
-			if (page_row>=Lines||cursor_row>=Lines)
-				 page_row=cursor_row=0;
-			if (page_row==Lines-1) {
-				/* so that it will not be thinked as a empty file */
-				page_row=Lines-page_len+1;
-				if (page_row<0) page_row=0;
-			}
-			pos_cur();
-			save_win_context(active_win);
-		}
-		if (file_no==1 && !batch) {
-			if (RestWinSize) setWinSize(fpos.wrow,fpos.wcol);
-			if (RestWinPos) setWinPos(fpos.wx,fpos.wy);
-		}
-	}
+  if (!getFileWininfo(fn,&fpos))
+    goto ll1;
+
+  if (RestCursor) {
+    page_row=fpos.page_row;
+    cursor_row=fpos.cursor_row;
+    cursor_col=fpos.cursor_col;
+    if (page_row>=Lines||cursor_row>=Lines)
+       page_row=cursor_row=0;
+    if (page_row==Lines-1) {
+      /* so that it will not be thinked as a empty file */
+      page_row=Lines-page_len+1;
+      if (page_row<0) page_row=0;
+    }
+  }
+#if	0
+  if (file_no==1 && !batch && !mult_open) {
+    if (RestWinSize) setWinSize(fpos.wrow,fpos.wcol);
+    if (RestWinPos) setWinPos(fpos.wx,fpos.wy);
+    fontidx=fpos.fntidx-1;
+    K_SwitchFont();
+  }
+#endif
+  save_win_context(active_win);
 }
+ll1:
 set_lang_type(fname);
+save_edstate();
 #if	0
 if (c_comment)
 	scan_comment(0,Lines);
@@ -407,6 +451,7 @@ if (file_no > 1 && !strcmp(edstate[0].fname,"#noname") && edstate[0].Lines==1 &&
 !edstate[0].edbf[0].len) {
 	close_file_id(0);
 }
+return 0;
 }
 
 void goto_file(int f_no)
@@ -423,7 +468,6 @@ static void edit_new_file()
 int i;
 char *s;
 
-
 if (extend_edstate()) return;
 
 hide_cursor_w();
@@ -434,13 +478,41 @@ if(!s[0]) {
 }
 save_edstate();
 s=decode_dir(s);
+
+if (strchr(s,'?')||strchr(s,'*')) {
+  char tt[256],ee[256];
+  int ll=0;
+  FILE *fp;
+
+  strcat(strcpy(tt,"/bin/ls -1 "),s);
+  if ((fp=popen(tt,"r"))==NULL) {
+    error("cannot execute /bin/ls -1");
+    return;
+  }
+  while (!feof(fp)) {
+    int l;
+    fgets(ee,sizeof(ee),fp);
+    l=strlen(ee);
+    if (l && ee[l-1]=='\n') ee[--l]=0;
+    if (!l || ee[l-1]=='~' || l>2 && ee[l-1]=='o' && ee[l-2]=='.')
+      continue;
+    for(i=0;i<file_no;i++)
+      if (!strcmp(ee,edstate[i].fname)) break;
+    if (i<file_no) continue;
+    edit_file(ee);
+  }
+  pclose(fp);
+  return;
+}
+
 for(i=0;i<file_no;i++)
 if (!strcmp(s,edstate[i].fname)) {
 	goto_file(i);
 	return;
 }
-strcpy(fname,s);
-edit_file();
+if (edit_file(s)) {
+	return;
+}
 disp_win_title();
 }
 
@@ -454,8 +526,10 @@ bzero(&edstate[file_no],sizeof(edstate[0]));
 page_row=page_col=cursor_row=cursor_col=cur_y=cur_x=f_modified=0;
 /* insert_mode=1; */
 cur_file=file_no;
+#if	0
 strcpy(fname,cmd);
-read_file(1);
+#endif
+read_file(cmd,1);
 save_win_context(active_win);
 file_no++;
 redraw_twin();
@@ -515,7 +589,7 @@ new_one:
 	if (fp) {
 		FPOS fpos;
 		int idx=0,match=0;
-		struct timeval tv;
+		extern int fontidx;
 
 		bzero(&fpos,sizeof(FPOS));
 		if (!newposf)
@@ -534,10 +608,8 @@ new_one:
 		fpos.cursor_col=edwin[active_win].pg_pos[c_file].cursor_col;
 		fpos.wrow=XW_ROW;
 		fpos.wcol=XW_COL;
+		fpos.fntidx=fontidx;
 		getWinXY(&fpos.wx,&fpos.wy);
-#if	0
-		gettimeofday(&tv,NULL);
-#endif
 		endian_conv((char *)&fpos.page_row);
 		endian_conv((char *)&fpos.cursor_row);
 		endian_conv((char *)&fpos.cursor_col);
@@ -545,9 +617,7 @@ new_one:
 		endian_conv((char *)&fpos.wcol);
 		endian_conv((char *)&fpos.wx);
 		endian_conv((char *)&fpos.wy);
-#if	0
-		endian_conv((char *)&tv.tv_sec);
-#endif
+		endian_conv((char *)&fpos.fntidx);
 		fwrite(&fpos,sizeof(FPOS),1,fp);
 		fclose(fp);
 	}
@@ -583,7 +653,10 @@ if (c_file < hot_file) hot_file--;
 if (c_file==hot_file) hot_file=-1;
 if (cur_file>=file_no) cur_file=file_no-1;
 edwin[active_win].cur_file=cur_file;
+restore_edstate();
+#if	0
 strcpy(fname,edstate[cur_file].fname);
+#endif
 set_lang_type(fname);
 redraw_twin();
 disp_win_title();
@@ -591,6 +664,7 @@ disp_win_title();
 
 void close_file()
 {
+save_edstate();
 close_file_id(cur_file,1);
 }
 
@@ -982,8 +1056,10 @@ if (i<file_no) {
 	return 0;
 }
 save_edstate();
+#if	0
 strcpy(fname,bf);
-edit_file();
+#endif
+edit_file(bf);
 #if	0
 page_row=row-(page_len>>1);
 #else
@@ -1071,7 +1147,7 @@ if (need_autosave && f_modified) {
 	strcat(strcpy(tt,fname),"#auto");
 	save_filename(tt,0);
 	message(5,"backup %s %d line ",tt,Lines);
-	show_cursor();
+  disp_cursor();
 	need_autosave=0;
 }
 }
@@ -1097,7 +1173,7 @@ for(row=0;row<Lines;row++)
 	}
 free(edbf);
 hide_cursor_w();
-read_file(0);
+read_file(fname,0);
 if (cursor_row>=Lines) cursor_row=Lines-1;
 if (cursor_col>edbf[cursor_row].len) cursor_col=edbf[cursor_row].len;
 disp_page();
@@ -1110,3 +1186,18 @@ void get_hostname()
 gethostname(hostname,sizeof(hostname));
 strcpy(user_name, (getpwuid(getuid()))->pw_name);
 }
+
+switch_fullname(char *s)
+{
+int i;
+char ff[128];
+
+for(i=0;i<file_no;i++) {
+	strcat(strcat(strcpy(ff,edstate[i].fdir),"/"),edstate[i].fname);
+	if (!strcmp(s,ff)) {
+		goto_file(i);
+		return 'Y';
+	}
+}
+return 0;
+}
