@@ -12,14 +12,18 @@
 #include <string.h>
 #include <X11/keysym.h>
 #include <X11/cursorfont.h>
+#include <X11/Xatom.h>
 #ifdef	AIX
 #include <sys/select.h>
 #endif
 
 #include "ted.h"
 
+static char *ted_atom="TedFname";
+static char *ted_res="TedRes";
+
 static char *ted_name="Ted";
-static int screen;
+static int screen_no;
 static Colormap colormap;
 static XColor mouse_fg_color,mouse_bg_color;
 static XFontStruct     *afont, *cfont;
@@ -37,9 +41,9 @@ int MCOL=M_COL;
 int X_inited=0;
 extern int batch;
 
-Window main_win;
+Window main_win,root;
 
-static Atom wm_del_win;
+Atom wm_del_win,atom_fname,atom_res;
 
 #define MARGIN 0
 
@@ -86,6 +90,10 @@ char *AscFontName="vga12x24";
 char *AscFontName="10x20";
 #endif
 
+int fontidx=-1;
+int fontN=5;
+char *fontlist[16];
+
 #include "color.h"
 
 ColorDef pc_color[13+MaxLang*(MaxKeyWType+8)]= {
@@ -107,7 +115,7 @@ ColorDef pc_color[13+MaxLang*(MaxKeyWType+8)]= {
 
 int ColorCnt=13;
 
-void check_alloc_color(int i, int fb)
+int check_alloc_color(int i, int fb)
 {
 int j;
 XColor color;
@@ -119,34 +127,39 @@ char *colorstr=(fb)?pc_color[i].bg_str:pc_color[i].fg_str;
 			if (fb) pc_color[i].bg=pc_color[j].fg;
 			else
 				pc_color[i].fg=pc_color[j].fg;
-			return;
+			return 0;
 		}
 		if (!strcmp(pc_color[j].bg_str,colorstr)) {
 			if (fb) pc_color[i].bg=pc_color[j].bg;
 			else
 				pc_color[i].fg=pc_color[j].bg;
-			return;
+			return 0;
 		}
 	}
 	if (!XParseColor(display,colormap,colorstr, &color))
 		p_err("invalid color %d %s",i, colorstr);
-	if (!XAllocColor(display,colormap,&color))
-		p_err("can't allocate color %s",pc_color[i].fg_str);
+	if (!XAllocColor(display,colormap,&color)) {
+		error("can't allocate color %s, will use private colormap.",
+		  pc_color[i].fg_str);
+		return -1;
+	}
 	if (fb)
 	pc_color[i].bg=color.pixel;
 	else
 	pc_color[i].fg=color.pixel;
+return 0;
 }
 
 static Cursor xcursor;
-void alloc_pc_colors()
+int alloc_pc_colors()
 {
 XColor color;
 int i,j;
 
 for(i=0;i<ColorCnt;i++)
-for(j=0;j<2;j++)
-	check_alloc_color(i,j);
+for(j=0;j<2;j++) {
+	if (check_alloc_color(i,j)<0) return -1;
+}
 foreground=pc_color[0].fg;
 background=pc_color[0].bg;
 if (!XParseColor(display,colormap,pc_color[11].fg_str, &mouse_fg_color))
@@ -157,34 +170,36 @@ if (!X_inited) xcursor = XCreateFontCursor(display,XC_xterm);
 XRecolorCursor(display,xcursor,&mouse_fg_color,&mouse_bg_color);
 if (X_inited)
 	XDefineCursor(display,main_win,xcursor);
+return 0;
 }
 
 
-static void load_font( char *fnt_name
-#if	CHINESE
-		,char *cfnt_name
-#endif
-)
+static void load_font(int setsize)
 {
-	int x, y, width, height;
-	int flags;
+int x, y, width, height;
+int flags;
 
-	if (!(afont=XLoadQueryFont(display, fnt_name)))
-	p_err("Cannot load ascii font %s", fnt_name);
+	if (afont) XFreeFont(display, afont);
+	setFontIndex();
+  if (!(afont=XLoadQueryFont(display, AscFontName)))
+		p_err("Cannot load ascii font %s", AscFontName);
 #if	CHINESE
-	if (!(cfont=XLoadQueryFont(display, cfnt_name)))
-	p_err("Cannot load Chinese font %s", cfnt_name);
+	if (cfont) XFreeFont(display, cfont);
+	if (!(cfont=XLoadQueryFont(display, Big5FontName)))
+ 	  p_err("Cannot load Chinese font %s", Big5FontName);
 #endif
 	fwidth=sizehints.width_inc = afont->max_bounds.width;
 	fheight=sizehints.height_inc = afont->ascent + afont->descent;
+	if (!setsize) return;
+
 	flags = XParseGeometry(geomstr,&x,&y,&width,&height);
 	if (flags & WidthValue) {
-	sizehints.flags |= USSize;
-	XW_COL=MCOL=width;
+		sizehints.flags |= USSize;
+		XW_COL=MCOL=width;
 	} else width=MCOL;
 	if (flags & HeightValue) {
-	sizehints.flags |= USSize;
-	XW_ROW=MROW=height;
+		sizehints.flags |= USSize;
+		XW_ROW=MROW=height;
 	} else height=MROW;
 	page_len=MROW-2;
 	page_width=MCOL;
@@ -198,7 +213,7 @@ static void load_font( char *fnt_name
 	sizehints.max_height=80* sizehints.height_inc;
 	if (flags & XValue) {
 	  if (flags & XNegative) {
-	  x = DisplayWidth(display,screen) + x - sizehints.width - 2;
+	  x = DisplayWidth(display,screen_no) + x - sizehints.width - 2;
 	  sizehints.win_gravity = NorthEastGravity;
 	}
 	  sizehints.x = x;
@@ -206,16 +221,86 @@ static void load_font( char *fnt_name
 	}
 	if (flags & YValue) {
 	  if (flags & YNegative) {
-	  y = DisplayHeight(display,screen) + y - sizehints.height - 2;
-	  sizehints.win_gravity = SouthWestGravity;
-	  if((flags&XValue)&&(flags&XNegative))
-	    sizehints.win_gravity = SouthEastGravity;
-	}
+			y = DisplayHeight(display,screen_no) + y - sizehints.height - 2;
+			sizehints.win_gravity = SouthWestGravity;
+			if((flags&XValue)&&(flags&XNegative))
+				sizehints.win_gravity = SouthEastGravity;
+	  }
 	  sizehints.y = y;
 	  sizehints.flags |= USPosition;
 	}
+
 }
 
+void create_gc()
+{
+XGCValues gcv;
+gcv.foreground = foreground;
+gcv.background = background;
+gcv.graphics_exposures = False;
+gcv.font=afont->fid;
+
+if (gc) XFreeGC(display,gc);
+gc = XCreateGC(display,main_win,GCForeground|GCBackground|GCFont|
+GCGraphicsExposures,&gcv);
+#if	CHINESE
+if (cgc) XFreeGC(display,gc);
+gcv.font=cfont->fid;
+cgc = XCreateGC(display,main_win,GCForeground|GCBackground|GCFont,&gcv);
+#endif
+}
+
+void setWinSize();
+
+void K_SwitchFont()
+{
+fontidx=(fontidx+1)%fontN;
+if (fontidx<0) {
+  fontidx=0;
+  return;
+}
+if (!strcmp(AscFontName,fontlist[fontidx]))
+  return;
+AscFontName=fontlist[fontidx];
+load_font(0);
+create_gc();
+setWinSize(XW_ROW,XW_COL);
+}
+
+setFontIndex()
+{
+int i;
+if (fontidx>=0) return;
+for(i=0;i<fontN;i++)
+  if (fontlist[i] && !strcmp(AscFontName,fontlist[i])) {
+    fontidx=i;
+    return;
+  }
+}
+
+void setFontList(char *s)
+{
+int i;
+char *t,*e=s+strlen(s);
+
+if (fontN) {
+	for(i=0;i<fontN;i++) free(fontlist[i]);
+}
+fontN=0;
+while (s<e) {
+	t=s;
+	while (*t!=' ' && *t!='\t' && *t) t++;
+	*t=0;
+	fontlist[fontN++]=strdup(s);
+	s=t+1;
+	while (s < e && *s==' ' || *s=='\t') s++;
+}
+#if	0
+printf("fontN:%d\n",fontN);
+for(i=0;i<fontN;i++)
+	puts(fontlist[i]);
+#endif
+}
 
 #define MW_EVENTS (	KeyPressMask |\
 			FocusChangeMask |\
@@ -234,12 +319,21 @@ static void create_win(int argc, char **argv)
 {
 	XWMHints wmhints;
 	XTextProperty name;
-
+	XSetWindowAttributes watt;
+#if	0
 	main_win = XCreateSimpleWindow(display,DefaultRootWindow(display),
-				 sizehints.x,sizehints.y,
+         sizehints.x,sizehints.y,
 				 sizehints.width,sizehints.height,
-				 1,foreground,background);
-
+         1,foreground,background);
+#else
+  watt.colormap=colormap;
+  watt.background_pixel=background;
+  watt.cursor=xcursor;
+	main_win = XCreateWindow(display,root,
+         sizehints.x,sizehints.y, sizehints.width,sizehints.height,
+         1,CopyFromParent,InputOutput,CopyFromParent,
+         CWBackPixel|CWColormap|CWCursor,&watt);
+#endif
 	class.res_name = ted_name;
 	class.res_class = "Ted";
 	wmhints.input = True;
@@ -255,19 +349,21 @@ static void create_win(int argc, char **argv)
 	XFree(name.value);
 
 	XSelectInput(display,main_win,MW_EVENTS);
-	XDefineCursor(display,main_win,xcursor);
-}
-
-static XErrorHandler XInErrorHandler(Display *dpy, XErrorEvent *event)
-{
-	p_err("Warning, error event receieved\n");
 #if	0
-	(void) XmuPrintDefaultErrorMessage (dpy, event, stderr);
+	XDefineCursor(display,main_win,xcursor);
 #endif
 }
 
-#define MAX_ROW (M_ROW+60)
-#define MAX_COL (M_COL+120)
+static XErrorHandler XTedErrorHandler(Display *dpy, XErrorEvent *event)
+{
+#if	0
+	(void) XmuPrintDefaultErrorMessage (dpy, event, stderr);
+exit(-1);
+#endif
+}
+
+#define MAX_ROW (M_ROW+100)
+#define MAX_COL (M_COL+200)
 static u_char tch[MAX_ROW][MAX_COL];	/* store text */
 static u_char tat[MAX_ROW][MAX_COL];	/* color attribute */
 static u_char tchd[MAX_ROW][MAX_COL];	/* store text */
@@ -341,30 +437,140 @@ for(i=0;i<len;i++)
 }
 #endif
 
+int cursor_x=0, cursor_y=0;
+
+
+void draw_edit_cursor(u_char att)
+{
+int x,y;
+int len;
+GC tgc;
+u_long fg,bg;
+
+	y=cursor_y;
+	x=cursor_x;
+	tatd[y][x]&=0xc0;
+	tatd[y][x]|=att;
+	if (tat[y][x] & 0x40) {
+	if (att!=1) {
+		x--;
+		goto xxxx;
+	}
+		XSetForeground(display, gc, pc_color[att].bg);
+	XDrawRectangle(display,main_win,gc,
+			x*fwidth,y*fheight,fwidth-1,fheight-1);
+	return;
+	}
+xxxx:
+#if	CHINESE
+	if (tat[y][x]&0x80 && (x<XW_COL-1) && (tat[y][x+1]&0x40)) {
+	len=2;
+		tatd[y][x+1]&=0xe0;
+		tatd[y][x+1]|=att;
+	}
+	else
+#endif
+	len=1;
+#if	CHINESE
+	tgc=(len==2?cgc:gc);
+#else
+	tgc=gc;
+#endif
+	XSetForeground(display, tgc, pc_color[att].fg);
+	XSetBackground(display, tgc, pc_color[att].bg);
+#if	CHINESE
+	if (len==2)
+	XDrawImageString16(display, main_win, tgc, x*fwidth,
+	y*fheight+cfont->ascent,(XChar2b *) &tch[y][x], 1);
+	else
+#endif
+	XDrawImageString(display, main_win, tgc, x*fwidth,
+	y*fheight+afont->ascent, &tch[y][x], 1);
+}
+
+void hide_cursor()
+{
+if (in_mlines(cursor_row) || in_mblock(cursor_row,page_col+cur_x)) {
+	draw_edit_cursor(3);
+}
+else {
+	draw_edit_cursor(0);
+}
+}
+
+int usexcopy=1;
 
 void draw_scr()
 {
 int y;
 u_char catt,cbyte;
 u_char *a,*b,*c,*d;
-int stx,i;
-int len;
+int stx,i,j,msti[MAX_ROW*MAX_ROW],mstj[MAX_ROW*MAX_ROW],mlen[MAX_ROW*MAX_ROW];
+int len,mcou;
 GC tgc;
 u_long fg,bg;
+
 #if	0
 dump1();
 dump2();
 #endif
 
 if (batch) return;
+mcou=0;
+if (!usexcopy) goto nocopy;
+for(i=0;i<XW_ROW;i++)
+	for(j=0;j<XW_ROW;j++)
+		if (i!=j &&
+		    !memcmp(&tch[i],&tchd[j],XW_COL) &&
+				!memcmp(&tat[i],&tatd[j],XW_COL)) {
+		  int k,l;
+		  k=i;
+		  l=j;
+			while (k < XW_ROW && l < XW_ROW &&
+			  !memcmp(&tch[k],&tchd[l],XW_COL) &&
+				!memcmp(&tat[k],&tatd[l],XW_COL)) {
+				k++;
+				l++;
+			}
+			len=k-i;
+			if (len) {
+			  msti[mcou]=i;
+			  mstj[mcou]=j;
+				mlen[mcou++]=len;
+				if (len > XW_ROW>>1) goto ll1;
+			}
+		}
+ll1:
+if (mcou) {
+	int cmidx=0,cmlen=0;
 
+	for(i=1;i<mcou;i++)
+		if (mlen[i] > mlen[cmidx])
+		  cmidx=i;
+
+	i=msti[cmidx];  /* new */
+	j=mstj[cmidx];  /* old */
+	len=mlen[cmidx];
+#if	0
+	printf("%d %d %d\n",i,j,len);
+#endif
+	for(y=i;y<i+len;y++) {
+		memcpy(&tchd[y][0],&tch[y][0],XW_COL);
+	  memcpy(&tatd[y][0],&tat[y][0],XW_COL);
+	}
+#if	1
+	XCopyArea(display,main_win,main_win,gc, 0,j*fheight,
+		XW_COL*fwidth,len*fheight, 0,i*fheight);
+#endif
+}
+
+nocopy:
 for(y=0;y<XW_ROW;y++) {
 	a=&tch[y][0];
 	b=&tchd[y][0];
 	c=&tat[y][0];
 	d=&tatd[y][0];
 	stx=0;
-
 	while (stx < XW_COL) {
 #if	CHINESE
 	while (*a==*b && !(*c & 0xc0) && *c==*d && stx<XW_COL) {
@@ -577,7 +783,6 @@ if (ctyp==2) costrd(strp-s+x,y,strp,cl,att);
 #endif
 }
 
-int cursor_x=0, cursor_y=0;
 u_char cur_att;
 void xprintf(char *fmt, ...)
 {
@@ -726,53 +931,6 @@ void clrscr_all()
 }
 
 
-void draw_edit_cursor(u_char att)
-{
-int x,y;
-int len;
-GC tgc;
-u_long fg,bg;
-
-	y=cursor_y;
-	x=cursor_x;
-	tatd[y][x]&=0xc0;
-	tatd[y][x]|=att;
-	if (tat[y][x] & 0x40) {
-	if (att!=1) {
-		x--;
-		goto xxxx;
-	}
-		XSetForeground(display, gc, pc_color[att].bg);
-	XDrawRectangle(display,main_win,gc,
-			x*fwidth,y*fheight,fwidth-1,fheight-1);
-	return;
-	}
-xxxx:
-#if	CHINESE
-	if (tat[y][x]&0x80 && (x<XW_COL-1) && (tat[y][x+1]&0x40)) {
-	len=2;
-		tatd[y][x+1]&=0xe0;
-		tatd[y][x+1]|=att;
-	}
-	else
-#endif
-	len=1;
-#if	CHINESE
-	tgc=(len==2?cgc:gc);
-#else
-	tgc=gc;
-#endif
-	XSetForeground(display, tgc, pc_color[att].fg);
-	XSetBackground(display, tgc, pc_color[att].bg);
-#if	CHINESE
-	if (len==2)
-	XDrawImageString16(display, main_win, tgc, x*fwidth,
-	y*fheight+cfont->ascent,(XChar2b *) &tch[y][x], 1);
-	else
-#endif
-	XDrawImageString(display, main_win, tgc, x*fwidth,
-	y*fheight+afont->ascent, &tch[y][x], 1);
-}
 
 void flush_x()
 {
@@ -799,32 +957,17 @@ for(i=page_len-1;i>top;i--)
 }
 }
 
-void show_cursor()
-{
-	draw_edit_cursor(1);
-}
-
 static int cur_tatt;
 void show_cursor_w()
 {
-if (batch) return;
-cursor_x=cur_x+win_x;
-cursor_y=cur_y+win_y;
-if (tat[cursor_y][cursor_x]!=1) cur_tatt=tat[cursor_y][cursor_x];
-if (insert_mode)
-	draw_edit_cursor(1);
-else
-	draw_edit_cursor(12);
-}
-
-void hide_cursor()
-{
-	if (in_mlines(cursor_row) || in_mblock(cursor_row,page_col+cur_x)) {
-		draw_edit_cursor(3);
-	}
-	else {
-		draw_edit_cursor(0);
-	}
+  if (batch) return;
+  cursor_x=cur_x+win_x;
+  cursor_y=cur_y+win_y;
+  if (tat[cursor_y][cursor_x]!=1) cur_tatt=tat[cursor_y][cursor_x];
+  if (insert_mode)
+    draw_edit_cursor(1);
+  else
+    draw_edit_cursor(12);
 }
 
 
@@ -851,6 +994,15 @@ else
 	draw_edit_cursor(12);
 }
 
+void disp_cursor()
+{
+if (!input_handler)
+ show_cursor_w();
+else
+if (input_handler<2)
+ show_cursor_r();
+}
+
 void hide_cursor_r(int color)
 {
 	draw_edit_cursor(color);
@@ -859,7 +1011,11 @@ void hide_cursor_r(int color)
 void inactive_cursor()
 {
 	hide_cursor();
-		XSetForeground(display, gc, pc_color[1].bg);
+	if (input_handler>1) return;
+  if (insert_mode && !input_handler || insert_mode_r && input_handler==1)
+    XSetForeground(display, gc, pc_color[1].bg);
+	else
+    XSetForeground(display, gc, pc_color[12].bg);
 	XDrawRectangle(display,main_win,gc,
 			cursor_x*fwidth,cursor_y*fheight,fwidth-1,fheight-1);
 }
@@ -900,72 +1056,70 @@ unsigned B2nd(int y, int x)
 }
 #endif
 
-static void restore();
 int visual_bell=0;
 void bell()
 {
-if (visual_bell) {
-	XClearWindow(display, main_win);
-	XFlush(display);
-	restore();
-}
-else
-XBell(display, 100);
+  if (visual_bell) {
+    XClearWindow(display, main_win);
+    XFlush(display);
+    restore();
+  }
+  else
+  XBell(display, 100);
 }
 
 extern int input_handler;
 
 void resize(int width, int height)
 {
-int nXW_COL=width/fwidth;
-int nXW_ROW=height/fheight;
-if (nXW_COL==XW_COL && nXW_ROW==XW_ROW) return;
-
-XW_COL=nXW_COL;
-XW_ROW=nXW_ROW;
-save_win_context(active_win);
-resize_twin();
-edln_restore();
-refresh_filebrowser();
-disp_kwc_pg();
-disp_kb_mode();
-if (!input_handler) {
-	show_cursor();
-} else
-if (input_handler==1) {
-	edln_restore();
-	show_cursor_r();
-}
+  int nXW_COL=width/fwidth;
+  int nXW_ROW=height/fheight;
+  if (nXW_COL==XW_COL && nXW_ROW==XW_ROW)
+    return;
+  XW_COL=nXW_COL;
+  XW_ROW=nXW_ROW;
+  save_win_context(active_win);
+  resize_twin();
+  edln_restore();
+  refresh_filebrowser();
+  disp_kwc_pg();
+  disp_kb_mode();
+  if (!input_handler) {
+    show_cursor_w();
+  } else
+  if (input_handler==1) {
+    edln_restore();
+    show_cursor_r();
+  }
 }
 
 
-static void restore()
+restore()
 {
 int y;
-	 for(y=0;y<XW_ROW;y++)
-	 	memset(tchd[y],0,XW_COL);
-	 draw_scr();
-	 if (!input_handler)
-	 	show_cursor();
-	 else
-	 	show_cursor_r();
+for(y=0;y<XW_ROW;y++)
+	bzero(tchd[y],XW_COL);
+
+draw_scr();
+disp_cursor();
 }
 
 static void scr_refresh(int x,int y, int width, int height)
 {
-int i,j;
-x/=fwidth;
-y/=fheight;
-width=(width+fwidth-1)/fwidth + 1;
-height=(height+fheight-1)/fheight + 1;
-for (j=y;j<y+height;j++)
-	memset(&tchd[j][x],0,width);
-draw_scr();
-if (!input_handler)
-	show_cursor();
-else {
-	if (input_handler!=3) show_cursor_r();
-}
+  int i,j;
+  int ousexcopy=usexcopy;
+
+  x/=fwidth;
+  y/=fheight;
+  width=(width+fwidth-1)/fwidth + 1;
+  height=(height+fheight-1)/fheight + 1;
+  for (j=y;j<y+height;j++) {
+    memset(&tchd[j][x],0,width);
+  }
+  usexcopy=0;
+  draw_scr();
+  usexcopy=ousexcopy;
+  disp_cursor();
 }
 
 
@@ -977,22 +1131,22 @@ Time eve_time;
 static int button0x=-1,button0y=-1,page_row0=-1;
 int CursorFollowClick=1;
 
+int xsockfd;
+
+#define REFRESH_PERIOD (2)
 #include <sys/time.h>
-void run_event()
+
+void X_event()
 {
 XEvent event;
 int x,y;
-int xsockfd=ConnectionNumber(display);
 fd_set chk_fd;
 struct timeval refreh_time;
 static int save_t_cou;
 static button_press_time;
 static motion_dire;
+extern int win_split;
 
-#define REFRESH_PERIOD (2)
-
-auto_save_time/=REFRESH_PERIOD;
-for(;;) {
 	XFlush(display);
 	FD_ZERO(&chk_fd);
 	FD_SET(xsockfd, &chk_fd);
@@ -1000,12 +1154,13 @@ for(;;) {
 	refreh_time.tv_usec=0;
 	if (!XPending(display) &&
 		!select(xsockfd+1,&chk_fd,NULL,NULL,&refreh_time)) {
-	show_time();
-	if (auto_save_time && save_t_cou>=auto_save_time) {
-		save_tmp_bak();
-		save_t_cou=0;
-	} else save_t_cou++;
-	continue;
+    show_time();
+    canCloseCwin();
+    if (auto_save_time && save_t_cou>=auto_save_time) {
+      save_tmp_bak();
+      save_t_cou=0;
+    } else save_t_cou++;
+    return;
 	}
 	while (XPending(display)) {
 	XNextEvent(display,&event);
@@ -1045,7 +1200,7 @@ for(;;) {
 	break;
 	case ClientMessage:
 	  if (event.xclient.format == 32 && event.xclient.data.l[0] == wm_del_win)
-		close_quit_file();
+		  close_quit_file();
 	break;
 	case SelectionClear:
 	break;
@@ -1064,13 +1219,15 @@ for(;;) {
 #if	1
 	case ConfigureNotify:
 	resize(event.xconfigure.width,event.xconfigure.height);
-#else
+	break;
+#endif
+#if	0
 	case ResizeRequest:
 	printf("wid:%d height:%d\n",event.xresizerequest.width,
 		event.xresizerequest.height);
 	resize(event.xresizerequest.width,event.xresizerequest.height);
-#endif
 	break;
+#endif
 	case ButtonPress:
 	if (input_handler) break;
 	x=event.xbutton.x;
@@ -1134,12 +1291,12 @@ for(;;) {
 	break;
 	case FocusIn:
 #if	CHINESE
-	send_FocusIn(display, main_win);
+	  send_FocusIn(display, main_win);
 #endif
-	show_cursor();
+    disp_cursor();
 	break;
 	case FocusOut:
-	inactive_cursor();
+	  inactive_cursor();
 	break;
 	case MotionNotify:
 	if (input_handler) break;
@@ -1197,12 +1354,60 @@ for(;;) {
 		break;
 	}
 	break;
+	case PropertyNotify:
+	  {
+			XPropertyEvent *ev=&event.xproperty;
+	    Window srcwin=ev->window;
+	    Atom atom,actual_type;
+	    int actual_format;
+			u_long nitems,bytes_after;
+			char *fname,res;
+
+	    if (ev->state==PropertyDelete || srcwin!=root || ev->atom!=atom_fname)
+	      break;
+			if (XGetWindowProperty(display,root,atom_fname, 0, 64,
+				False, AnyPropertyType, &actual_type,&actual_format,
+				&nitems,&bytes_after,(u_char **)&fname) != Success)
+				puts("err prop");
+			if (nitems==0) break;
+			res=switch_fullname(fname);
+			XFree(fname);
+			if (res=='Y') {
+				XChangeProperty(display,root, atom_res, XA_STRING, 8,
+						PropModeReplace, (u_char *)&res, 1);
+				XSync(display,False);
+			  XRaiseWindow(display,main_win);
+			}
+		}
+	break;
+	case VisibilityNotify:
+	  if(event.xvisibility.state == VisibilityUnobscured) {
+			usexcopy = 1;
+	  }
+	  else {
+		  usexcopy = 0;
+		}
+	break;
 #if	0
 	default:
 	printf("Unknow event %d\n",event.type);
 #endif
 	}
 	}  /* while XPend */
+
+}
+
+
+
+
+void run_event()
+{
+
+xsockfd=ConnectionNumber(display);
+
+auto_save_time/=REFRESH_PERIOD;
+for(;;) {
+	X_event();
 }
 
 }
@@ -1248,6 +1453,9 @@ void get_arg(int argc, char **argv)
 int num=sizeof(arg)/sizeof(ARG);
 char *pp;
 int i,j;
+FPOS fpos;
+extern int fnamesN;
+extern char *fnames[];
 
 if (!batch)
 for(i=0;i<num;i++)
@@ -1292,7 +1500,7 @@ for (i=1;i<argc;i++) {
 		}
 		if (j==num) {
 			char tt[128];
-			fprintf(stderr,"+==== Ted editor ver 4.2d ====+\n");
+			fprintf(stderr,"+==== Ted editor ver 4.4d ====+\n");
 			fprintf(stderr,"| by Edward D. H. Liu, Taiwan |\n");
 			fprintf(stderr,"| dhliu@solar.csie.ntu.edu.tw |\n");
 			fprintf(stderr,"+-----------------------------+\n");
@@ -1318,6 +1526,15 @@ next:
 /* HP-UX cc bug */
 	1;
 }
+
+if (fnamesN && getFileWininfo(fnames[0],&fpos)) {
+  char tt[80];
+  fontidx=fpos.fntidx;
+  AscFontName=fontlist[fontidx];
+  sprintf(tt,"%dx%d+%d+%d",fpos.wcol,fpos.wrow,fpos.wx,fpos.wy);
+  geomstr=strdup(tt);
+}
+
 }
 
 
@@ -1326,12 +1543,14 @@ void change_win_name(char *str)
 {
 	XTextProperty name;
 
-if (XStringListToTextProperty(&str,1,&name) == 0) {
-	error("change_window_name");
-	return;
-}
-XSetWMName(display,main_win,&name);
-XFree(name.value);
+  if (XStringListToTextProperty(&str,1,&name) == 0) {
+    error("change_window_name");
+    return;
+  }
+  XSetWMName(display,main_win,&name);
+  XStoreName(display,main_win,str);
+  XSetIconName(display,main_win,str);
+  XFree(name.value);
 }
 
 void getWinXY(int *x, int *y)
@@ -1343,43 +1562,40 @@ XTranslateCoordinates(display,main_win,DefaultRootWindow(display),0,0,x,y,
 
 void setWinSize(int row, int col)
 {
-if (row>0 && col>0)
-	XResizeWindow(display,main_win,col*fwidth,row*fheight);
+  if (row>0 && col>0) {
+    XResizeWindow(display,main_win,col*fwidth,row*fheight);
+  }
 }
 
-int first_time=1;
+void K_SetWinSize(row,col)
+{
+  setWinSize(col,row);
+}
+
+static int first_time=1;
 
 void setWinPos(int x, int y)
 {
-#if	0
-Window w,ow,root,parent,*chlist;
-int listn,count=0;
-
-if (x<=0 && y<=0) return;
-ow=w=main_win;
-do {
-	XQueryTree(display,w,&root,&parent,&chlist,&listn);
-	if (listn) XFree(chlist);
-	ow=w;
-	w=parent;
-	count++;
-} while (parent!=root && count < 10);
-
-XMoveWindow(display,ow,x,y);
-#else
 Window ow;
 int ox,oy;
 
 if (first_time) { /* somtimes receives X errors */
-	XSync(display,False);
-	usleep(50000);
+	int cou=0;
+	XWindowAttributes att;
+	/* waiting until main_win shows up */
+	do {
+		XGetWindowAttributes(display, main_win, &att);
+		if (att.map_state==IsViewable) break;
+		XSync(display,False);
+		usleep(50000);
+	} while (cou<100);
+	if (cou==100) return;
 	first_time=0;
 }
 if (x<=0 && y<=0) return;
 XTranslateCoordinates(display,main_win,DefaultRootWindow(display),0,0,&ox,&oy,
 &ow);
 XMoveWindow(display,ow,x,y);
-#endif
 }
 
 void init_X_arg(int argc, char **argv)
@@ -1396,7 +1612,7 @@ for(i=1;i<argc;i++)
 if (!batch) {
 	if (!(display = XOpenDisplay(display_name)))
 		p_err("Cannot open display %s", display_name);
-	XSetErrorHandler((XErrorHandler)XInErrorHandler);
+	XSetErrorHandler((XErrorHandler)XTedErrorHandler);
 }
 
 get_arg(argc,argv);
@@ -1409,35 +1625,135 @@ XWarpPointer(display,None,main_win,0,0,0,0,
 		fwidth*(XW_COL>>1),fheight*(XW_ROW>>1));
 }
 
+int owncmap=0;
+
 void init_X_win(int argc, char **argv)
 {
-XGCValues gcv;
+int our_depth;
+Visual *our_visual;
 
-colormap = DefaultColormap(display,screen);
-screen = DefaultScreen(display);
-alloc_pc_colors();
+
+root=DefaultRootWindow(display);
+screen_no = DefaultScreen(display);
+
+load_font(1);
+if (owncmap) {
+  XVisualInfo *list;
+  XVisualInfo template;
+  int nitems_return;
+  Screen *screen;
+  XVisualInfo *best,*list1;
+
+install_cmap:
+  best=NULL;
+  screen = DefaultScreenOfDisplay(display);
+  our_depth = DefaultDepthOfScreen(screen);
+  our_visual = DefaultVisualOfScreen(screen);
+
+  template.screen = screen_no;
+  template.class = TrueColor;
+  colormap = DefaultColormapOfScreen(screen);
+  list = XGetVisualInfo(display, VisualScreenMask | VisualClassMask,
+    &template, &nitems_return);
+
+  for (list1 = list; list1 < list + nitems_return; ++list1)
+    if (list1->depth > our_depth
+      && (best == NULL || list1->depth > best->depth))
+       best = list1;
+    if (list!=NULL) {
+      if (best!=NULL) {
+        our_depth = best->depth;
+        our_visual = best->visual;
+        colormap = XCreateColormap(display,root,our_visual,AllocAll);
+        colormap = XCopyColormapAndFree(display, colormap);
+      }
+      XFree(list);
+    }
+    if (our_visual->class == PseudoColor) {
+      XColor tmp_color;
+      u_long plane_masks[16],pixel;
+      colormap = XCreateColormap(display,root,our_visual,AllocNone);
+      alloc_pc_colors();
+#if	1
+	    colormap = XCopyColormapAndFree(display, colormap);
+#endif
+      XInstallColormap(display,colormap);
+	  } else
+    alloc_pc_colors();
+} else {
+  colormap = DefaultColormap(display,screen_no);
+  if (alloc_pc_colors()<0)
+    goto install_cmap;
+}
+create_win(argc,argv);
 
 bzero(tat, sizeof(tat));
-#if	CHINESE
-load_font(AscFontName,Big5FontName);
-#else
-load_font(AscFontName);
-#endif
-create_win(argc,argv);
 XDefineCursor(display,main_win,xcursor);
-XMapWindow(display,main_win);
 
+create_gc();
 set_att(0);
 wm_del_win = XInternAtom(display,"WM_DELETE_WINDOW",False);
 XSetWMProtocols(display,main_win,&wm_del_win,1);
-gcv.foreground = foreground;
-gcv.background = background;
-gcv.font=afont->fid;
-gc = XCreateGC(display,main_win,GCForeground|GCBackground|GCFont,&gcv);
-#if	CHINESE
-gcv.font=cfont->fid;
-cgc = XCreateGC(display,main_win,GCForeground|GCBackground|GCFont,&gcv);
-#endif
 memset(tchd,255,sizeof(tchd));
 X_inited=1;
+XSelectInput(display, root, PropertyChangeMask);
+atom_fname=XInternAtom(display,ted_atom,False);
+atom_res=XInternAtom(display,ted_res,False);
+init_xview();
+
+XMapWindow(display,main_win);
 }
+
+
+int mult_open,ExitDupOpen;
+int test_file(char *fdir,char *fname)
+{
+Window twin;
+Atom atom;
+Window win;
+char tfname[128];
+XClientMessageEvent event;
+Atom actual_type;
+int actual_format;
+u_long nitems,bytes_after;
+u_char *res;
+
+if (strcmp(fdir,"/")) {
+	strcpy(tfname,fdir);
+	strcat(strcat(tfname,"/"),fname);
+} else strcpy(tfname,fname);
+
+XChangeProperty(display, root , atom_res, XA_STRING, 8,
+		PropModeReplace, "    ",1);
+XSync(display,False);
+
+XChangeProperty(display, root , atom_fname, XA_STRING, 8,
+		PropModeReplace, (u_char *)tfname, strlen(tfname)+1);
+XSync(display,False);
+
+usleep(100000);
+
+if (XGetWindowProperty(display,root,atom_res, 0, 1,
+	False, AnyPropertyType, &actual_type,&actual_format,
+	&nitems,&bytes_after,&res) != Success)
+	puts("err prop");
+if (nitems==0) {
+	return 0;
+}
+
+if (*res=='Y') {
+	extern int readonly;
+	char *tmp=event.data.b;
+
+	message(4,"Already opened");
+	XSync(display,False);
+	if (ExitDupOpen && !file_no) {
+	  p_err("Dup open %s",tfname);
+	}
+	return 1;
+}
+
+XFree(res);
+return 0;
+}
+
