@@ -9,6 +9,7 @@
 #include <stdarg.h>
 #include <stdio.h>
 #include <string.h>
+#include <stdlib.h>
 #include "ted.h"
 #include "xkey.h"
 #include "color.h"
@@ -90,22 +91,28 @@ extern K_cur_upN(),K_cur_downN(),K_SaveNLtoPaste(),K_PasteBuffPutAfter();
 extern K_delvi(),K_SaveEoltoPaste(),K_PasteBuffer(),K_SaveWordtoPaste();
 extern K_WriteBlk(),K_GotoGrep(),K_Tab2Space(),K_TogVertical();
 extern K_LocateStr(),K_ReplaceStr(),K_ExportMacro(),K_TogReadOnly();
-extern K_VisitedFile(),K_quit2();
-
+extern K_VisitedFile(),K_quit2(),K_SwitchFont(),K_TogDispTab(),K_SetWinSize();
+extern K_ExportStatus(),K_SumRectBlock();
 /*
  field 1 :
  field 2 :
  field 3 : arg type
+           0 : no arg
+           1 : char     (integer)
+           2 : string
+           3 : string string
+           7 : string string string
+           8 : int int
  field 4 : If eq 1, this func will modify text
 */
 
 FUNCMAP func[]={
-{"PutStr",K_PutStr,2,    		1}, 		/* PutStr must be the 1st one */
-{"RecMacro",K_RecMacro,0}, 			/* fixed to 1 */
-{"PlayMacro",K_PlayMacro,0}, 		/* fixed to 2 */
+{"PutStr",K_PutStr,2,       1},     /* PutStr must be the 1st one */
+{"RecMacro",K_RecMacro,0},      /* fixed to 1 */
+{"PlayMacro",K_PlayMacro,0},    /* fixed to 2 */
 {"SaveCurFuncs",K_PlayMacro,5}, /*          3 */
 {"PlaySaveFuncs",K_PlayMacro,6},/*          4 */
-{"enter",K_ReturnIndent,0,	1},
+{"enter",K_ReturnIndent,0,  1},
 {"up",K_cur_up,0},
 {"down",K_cur_down,0},
 {"left",K_cur_left,0},
@@ -124,32 +131,32 @@ FUNCMAP func[]={
 {"quit2",K_quit2,0},
 {"bs",K_BS,0,1},
 {"del",K_del,0,1},
-{"DelLn",K_DelLn,0,				1},
-{"DelEol",K_DelEol,0,			1},
-{"DelWord",K_DelWord,0,		1},
-{"CpArea",K_CpArea,0,			1},
-{"SaveFile",K_SaveFile,0,	1},
+{"DelLn",K_DelLn,0,       1},
+{"DelEol",K_DelEol,0,     1},
+{"DelWord",K_DelWord,0,   1},
+{"CpArea",K_CpArea,0,     1},
+{"SaveFile",K_SaveFile,0, 1},
 {"SwWin", K_SwWin,0},
 {"PrevFile",K_PrevFile,0},
 {"NextFile",K_NextFile,0},
 {"MarkLn",K_MarkLn,0},
 {"MarkBlk",K_MarkBlk,0},
 {"Unmark",K_Unmark,0},
-{"DelArea",K_DelArea,0,		1},
-{"OwCp",K_OwCp,0,					1},
-{"MvArea",K_MvArea,0,			1},
+{"DelArea",K_DelArea,0,   1},
+{"OwCp",K_OwCp,0,         1},
+{"MvArea",K_MvArea,0,     1},
 {"EdFile",K_EdFile,0},
-{"TogIns",K_TogIns,0,			1},
+{"TogIns",K_TogIns,0,     1},
 {"SplitWin",K_SplitWin,0},
 {"FindStr",K_FindStr,0},
 {"RepStr",K_RepStr,0},
 {"WrFile",K_WrFile,0},
-{"Undo",K_Undo,0,					1},
-{"Redo",K_Redo,0,					1},
+{"Undo",K_Undo,0,         1},
+{"Redo",K_Redo,0,         1},
 {"system",K_system,2},
 {"SetTag",K_SetTag,1},
 {"GotoTag",K_GotoTag,1},
-{"JoinLine",K_JoinLine,0,		1},
+{"JoinLine",K_JoinLine,0,   1},
 {"ExtMarkUp",K_ExtMarkUp,0},
 {"ExtMarkDown",K_ExtMarkDown,0},
 {"ExtMarkLeft",K_ExtMarkLeft,0},
@@ -172,7 +179,7 @@ FUNCMAP func[]={
 {"TgHotFile",K_TgHotFile,0},
 {"InsPipe",K_InsPipe,2,1},
 {"LookupTag",K_LookupTag,0},
-{"ExecCmd",K_ExecCmd,0,1},
+{"ExecCmd",K_ExecCmd,0,0},
 {"CurLeftWord",K_CurLeftWord,0},
 {"CurRightWord",K_CurRightWord,0},
 {"UpLowBlk",K_UpLowBlk,1},
@@ -186,12 +193,12 @@ FUNCMAP func[]={
 {"TogIndent",K_TogIndent,0},
 {"ExeCmdFnameExt",K_ExeCmdFnameExt,3},
 {"DelBol",K_DelBol,0,1},
-{"BShiftUpDn",K_BShiftUpDn,1,		1},
-{"BShiftLR",K_BShiftLR,1,				1},
+{"BShiftUpDn",K_BShiftUpDn,1,   1},
+{"BShiftLR",K_BShiftLR,1,       1},
 {"MergeTedrc",K_MergeTedrc,2},
 {"LookMarkTag",K_LookMarkTag,0},
 {"FindMarkStr",K_FindMarkStr,0},
-{"Dup1stToEndBlk",K_Dup1stToEndBlk,0,		1},
+{"Dup1stToEndBlk",K_Dup1stToEndBlk,0,   1},
 {"MatchBracket", K_MatchBracket,0},
 {"FileBrowser", K_FileBrowser,0},
 {"PrevFind", K_PrevFind,0},
@@ -201,11 +208,11 @@ FUNCMAP func[]={
 {"ToPasteBuff",K_ToPasteBuff,0},
 {"PasteBuffPut",K_PasteBuffPut,0},
 {"PasteBuffPutAfter",K_PasteBuffPutAfter,0},
-{"DelLnN",K_DelLnN,1,										1},
+{"DelLnN",K_DelLnN,1,                   1},
 {"upN",K_cur_upN,1},
 {"downN",K_cur_downN,1},
 {"SaveNLtoPaste",K_SaveNLtoPaste,1},
-{"delvi",K_delvi,0,											1},
+{"delvi",K_delvi,0,                     1},
 {"SaveEoltoPaste",K_SaveEoltoPaste,0},
 {"PasteBuffer",K_PasteBuffer,0},
 {"SaveWordtoPaste",K_SaveWordtoPaste,0},
@@ -214,10 +221,15 @@ FUNCMAP func[]={
 {"Tab2Space",K_Tab2Space,0},
 {"TogVertical",K_TogVertical,0},
 {"LocateStr",K_LocateStr,3},
-{"ReplaceStr",K_ReplaceStr,7,						1},
+{"ReplaceStr",K_ReplaceStr,7,           1},
 {"ExportMacro",K_ExportMacro,2},
 {"TogReadOnly",K_TogReadOnly,0},
 {"VisitedFile",K_VisitedFile,2},
+{"SwitchFont",K_SwitchFont,0},
+{"TogDispTab",K_TogDispTab,0},
+{"SetWinSize",K_SetWinSize,8},
+{"ExportStatus",K_ExportStatus,2},
+{"SumRectBlock",K_SumRectBlock,0},
 };
 
 int funcno=sizeof(func)/sizeof(func[0]);
@@ -268,7 +280,8 @@ char *caddr;
 extern int auto_save_time;
 extern int wrap_width, LeadingSpaceTab,Indent,BackupSave,TrimBlankTail;
 extern int CurAutoUpDown,StripCR,CComp,WinSplitSeq,CursorFollowClick;
-extern int AutoBlockRepDel,RestCursor,RestWinPos,RestWinSize;
+extern int AutoBlockRepDel,RestCursor,RestWinPos,RestWinSize,ExitDupOpen;
+extern int FModifiedQuitYN,DispTab,TabChar;
 struct {
 char *sname;
 int *saddr;
@@ -291,6 +304,10 @@ int *saddr;
 {"RestCursor",&RestCursor},
 {"RestWinSize",&RestWinSize},
 {"RestWinPos",&RestWinPos},
+{"ExitDupOpen",&ExitDupOpen},
+{"FModifiedQuitYN",&FModifiedQuitYN},
+{"DispTab",&DispTab},
+{"TabChar",&TabChar},
 };
 
 char *tedrcfname="tedrc";
@@ -334,64 +351,134 @@ int hex2b(char h)
 return (h>='a'?h-'a'+10:h-'0');
 }
 
+void mkstr(char *aa,char *bb,char *tt)
+{
+  int len;
+  memcpy(tt,aa,len=bb-aa);
+  tt[len]=0;
+}
+
+char *varname(char *s,char *fn)
+{
+  char *t;
+  int len;
+  t=s=skip_spc(s);
+  if (!isvarchr(*s)) return 0;
+  while (isvarchr(*t)||isdig(*t)) t++;
+  mkstr(s,t,fn);
+  return t;
+}
+
+
+char *varname_sh(char *s,char *fn)
+{
+  char *t;
+  int len;
+  char uu[256];
+  char *env;
+
+  t=s+1;
+  if (!isvarchr(*t)) {
+skip:
+    strcpy(fn,"$");
+    return s+1;
+  }
+  while (isvarchr(*t)||isdig(*t))
+    t++;
+  mkstr(s+1,t,uu);
+  if (env=getenv(uu)) {
+    strcpy(fn,env);
+    return t;
+  } else {
+    if (!strcmp(uu,"TEDDIR")) {
+      strcpy(fn,TEDDIR);
+      return t;
+    } else
+    if (!strcmp(uu,"DBGSTOP")) {
+#ifdef	__GNUC__
+      strcpy(fn,"b");
+#else
+      strcpy(fn,"stop at");
+#endif
+      return t;
+    }
+    else
+      goto skip;
+  }
+}
+
+
 char *parse_Cstr(char *tedname, int lno, char *ss, char **end)
 {
-static char tt[1024];
-int sl=0;
-if (*ss!='"') lper(tedname,lno,"Bad string %s",ss);
-ss++;
-while (*ss && *ss!='"') {
-	if (*ss=='\\') {
-		switch (*(ss+1)) {
-			case 'n':
-				tt[sl++]='\n';
-				ss+=2;
-				continue;
-			case 'r':
-				tt[sl++]='\r';
-				ss+=2;
-				continue;
-			case '\\':
-				tt[sl++]='\\';
-				ss+=2;
-				continue;
-			case 't':
-				tt[sl++]='\t';
-				ss+=2;
-				continue;
-			case '"':
-				tt[sl++]='"';
-				ss+=2;
-				continue;
-			case 'v':
-				tt[sl++]='\v';
-				ss+=2;
-				continue;
-			case 'f':
-				tt[sl++]='\f';
-				ss+=2;
-				continue;
-			case 'a':
-				tt[sl++]='\a';
-				ss+=2;
-				continue;
-			case 'e':
-				tt[sl++]=27;
-				ss+=2;
-				continue;
-			case 'x':
-				ss+=2;
-				tt[sl++]=(hex2b(*ss)<<4)+hex2b(*(ss+1));
-				ss+=2;
-				continue;
-		}
-	}
-	tt[sl++]=*(ss++);
-}
-if (!(*ss)) lper(tedname,lno,"\" expected");
-tt[sl]=0;
-*end=ss+1;
-return tt;
+  static char tt[1024];
+  int sl=0;
+  if (*ss!='"') lper(tedname,lno,"Bad string %s",ss);
+  ss++;
+  while (*ss && *ss!='"') {
+    if (*ss=='\\') {
+      switch (*(ss+1)) {
+        case 'n':
+          tt[sl++]='\n';
+          ss+=2;
+          continue;
+        case 'r':
+          tt[sl++]='\r';
+          ss+=2;
+          continue;
+        case '\\':
+          tt[sl++]='\\';
+          ss+=2;
+          continue;
+        case 't':
+          tt[sl++]='\t';
+          ss+=2;
+          continue;
+        case '"':
+          tt[sl++]='"';
+          ss+=2;
+          continue;
+        case 'v':
+          tt[sl++]='\v';
+          ss+=2;
+          continue;
+        case 'f':
+          tt[sl++]='\f';
+          ss+=2;
+          continue;
+        case 'a':
+          tt[sl++]='\a';
+          ss+=2;
+          continue;
+        case 'e':
+          tt[sl++]=27;
+          ss+=2;
+          continue;
+        case 'x':
+          ss+=2;
+          tt[sl++]=(hex2b(*ss)<<4)+hex2b(*(ss+1));
+          ss+=2;
+          continue;
+        case '$':
+          tt[sl++]='$';
+          ss+=2;
+          continue;
+      }
+    } else
+    if (*ss=='$') {
+      char uu[1024];
+      int l;
+      ss=varname_sh(ss,uu);
+      l=strlen(uu);
+      memcpy(&tt[sl],uu,l);
+      sl+=l;
+      continue;
+    }
+    tt[sl++]=*(ss++);
+  }
+  if (!(*ss)) lper(tedname,lno,"\" expected");
+  tt[sl]=0;
+  *end=ss+1;
+  return tt;
 }
 
 static parse_key(char *tedname, int lno, char *uu,char *vv,u_short *oo)
@@ -401,50 +488,50 @@ char *ww,*xx,*yy;
 u_short kmod,kname;
 
 while (*uu && uu<vv) {
-	if (*uu!='[') lper(tedname,lno,"cannot find the 1st [");
-	xx=++uu;
-	kmod=0;
-	yy=xx;
-	while (*yy && *yy!='-' && *yy!=']') yy++;
-	while (*xx!='-' && *xx!=']' && xx<vv) {
-		if (*yy=='-')
-		switch (*xx) {
-		case 'c':
-			kmod|=CTRL;
-			break;
-		case 'a':
-			kmod|=ALT;
-			break;
-		case 's':
-			kmod|=SHFT;
-			break;
-		}
-		xx++;
-	}
-	if (!kmod) {
-		if (*xx==']') {
-			ww=xx;
-			*xx=0;
-		}
-		else lper(tedname,lno,"cannot find ]");
-		xx=uu;
-	}
-	else
-	if (*xx=='-') {
-		ww=++xx;
-		while (*ww!=']' && ww<vv) ww++;
-		if (*ww!=']') lper(tedname,lno,"cannot find ]");
-		else *ww=0;
-	} else lper(tedname,lno,"error");
-	for(i=0;i<kname_no;i++)
-	if (!strcmp(xx,kmap[i].keyname)){
-		kname=i;
-		break;
-	}
-	if (i==kname_no) lper(tedname,lno,"unknown key %s",xx);
-	kname|=kmod;
-	oo[tk++]=kname;
-	uu=ww+1;
+  if (*uu!='[') lper(tedname,lno,"cannot find the 1st [");
+  xx=++uu;
+  kmod=0;
+  yy=xx;
+  while (*yy && *yy!='-' && *yy!=']') yy++;
+  while (*xx!='-' && *xx!=']' && xx<vv) {
+    if (*yy=='-')
+    switch (*xx) {
+    case 'c':
+      kmod|=CTRL;
+      break;
+    case 'a':
+      kmod|=ALT;
+      break;
+    case 's':
+      kmod|=SHFT;
+      break;
+    }
+    xx++;
+  }
+  if (!kmod) {
+    if (*xx==']') {
+      ww=xx;
+      *xx=0;
+    }
+    else lper(tedname,lno,"cannot find ]");
+    xx=uu;
+  }
+  else
+  if (*xx=='-') {
+    ww=++xx;
+    while (*ww!=']' && ww<vv) ww++;
+    if (*ww!=']') lper(tedname,lno,"cannot find ]");
+    else *ww=0;
+  } else lper(tedname,lno,"error");
+  for(i=0;i<kname_no;i++)
+  if (!strcmp(xx,kmap[i].keyname)){
+    kname=i;
+    break;
+  }
+  if (i==kname_no) lper(tedname,lno,"unknown key %s",xx);
+  kname|=kmod;
+  oo[tk++]=kname;
+  uu=ww+1;
 }
 return tk;
 }
@@ -469,55 +556,38 @@ ispc(char c)
 return (c==' ' || c==9);
 }
 
-void mkstr(char *aa,char *bb,char *tt)
-{
-int len;
-memcpy(tt,aa,len=bb-aa);
-tt[len]=0;
-}
-
-char *varname(char *s,char *fn)
-{
-char *t;
-int len;
-t=s=skip_spc(s);
-if (!isvarchr(*s)) return 0;
-while (isvarchr(*t)||isdig(*t)) t++;
-mkstr(s,t,fn);
-return t;
-}
 
 u_char takearg(char *tedname, int lno, char *ss, char **ostr, char **next)
 {
-static char ooo[64];
-char *tt;
-char *aa;
-ss=skip_spc(ss);
-*next=tt=ss;
-if (!*ss) return 0;
-if (isdig(*tt)) {
-	while (isdig(*tt)) tt++;
-	mkstr(ss,*next=tt,*ostr=ooo);
-	return 1;
-}
-else
-if (*tt=='"') {
-	*ostr=parse_Cstr(tedname, lno, tt,next);
-	return 2;
-}
-return 0;
+  static char ooo[64];
+  char *tt, *env;
+  char *aa;
+  ss=skip_spc(ss);
+  *next=tt=ss;
+  if (!*ss) return 0;
+  if (isdig(*tt)) {
+    while (isdig(*tt)) tt++;
+    mkstr(ss,*next=tt,*ostr=ooo);
+    return 1;
+  }
+  else
+  if (*tt=='"') {
+    *ostr=parse_Cstr(tedname, lno, tt,next);
+    return 2;
+  }
+  return 0;
 }
 
 char *getstr(char *s, char **t)
 {
-static char tt[80];
-char *u;
-while (ispc(*s) || *s==',') s++;
-u=s;
-while (isvarchr(*u) || u>s && isdigit(*u)) u++;
-mkstr(s,u,tt);
-*t=u;
-return tt;
+  static char tt[80];
+  char *u;
+  while (ispc(*s) || *s==',') s++;
+  u=s;
+  while (isvarchr(*u) || u>s && isdigit(*u)) u++;
+  mkstr(s,u,tt);
+  *t=u;
+  return tt;
 }
 
 parse_stmt(char *tedname, int lno,  char *ss,char **cmd, u_char *count)
@@ -525,82 +595,99 @@ parse_stmt(char *tedname, int lno,  char *ss,char **cmd, u_char *count)
 char fn[32], *arg,*tt;
 char tbuf[1024];
 u_short funidx[64],fuid;
-int typ,cmlen=0,fncnt=0,argtyp;
+int typ,cmlen=0,fncnt=0,argtyp,ii;
 while (*ss) {
-	if (!(ss=varname(ss,fn))) lper(tedname,lno,"func name expected");
-	for(fuid=0;fuid<funcno;fuid++)
-		if (!strcmp(fn,func[fuid].funname))
-			break;
-	if (fuid==funcno) {
-		lper2(tedname,lno,"func %s is not defined",fn);
-		return 0;
-	}
-	ss=skip_spc(ss);
-	if (*ss!='(') lper(tedname,lno,"( expected");
-	ss++;
-	funidx[fncnt++]=cmlen;
-	argtyp=func[fuid].argtype;
-	tbuf[cmlen++]=fuid;
-	switch (argtyp) {
-	case 0:
-	case 4:
-	case 5:
-	case 6:
-		typ=takearg(tedname, lno, ss,&arg,&tt);
-		if (typ!=0)
-			lper(tedname,lno,"No arg is expected");
-		break;
-	case 1:
+  if (!(ss=varname(ss,fn))) lper(tedname,lno,"func name expected");
+  for(fuid=0;fuid<funcno;fuid++)
+    if (!strcmp(fn,func[fuid].funname))
+      break;
+  if (fuid==funcno) {
+    lper2(tedname,lno,"func %s is not defined",fn);
+    return 0;
+  }
+  ss=skip_spc(ss);
+  if (*ss!='(') lper(tedname,lno,"( expected");
+  ss++;
+  funidx[fncnt++]=cmlen;
+  argtyp=func[fuid].argtype;
+  tbuf[cmlen++]=fuid;
+  switch (argtyp) {
+  case 0:
+  case 4:
+  case 5:
+  case 6:
+    typ=takearg(tedname, lno, ss,&arg,&tt);
+    if (typ!=0)
+      lper(tedname,lno,"No arg is expected");
+    break;
+  case 1:
+    typ=takearg(tedname, lno, ss,&arg,&tt);
+    if (typ!=1)
+      lper(tedname, lno, tedname,lno,"integer is expected");
+    tbuf[cmlen++]=atoi(arg);
+    break;
+  case 2:
+    typ=takearg(tedname, lno, ss,&arg,&tt);
+    if (typ!=2)
+      lper(tedname,lno,"String is expected");
+    strcpy(&tbuf[cmlen],arg);
+    cmlen+=strlen(arg)+1;
+    break;
+  case 3:  /* (string,string) */
+    typ=takearg(tedname, lno,ss,&arg,&tt);
+    if (typ!=2)
+      lper(tedname,lno,"String is expected");
+    strcpy(&tbuf[cmlen],arg);
+    cmlen+=strlen(arg)+1;
+    tt=skip_spc(tt);
+    if (*tt!=',') lper(tedname,lno,", is expected");
+    tt++;
+    typ=takearg(tedname, lno,tt,&arg,&tt);
+    if (typ!=2)
+      lper(tedname,lno,"String is expected");
+    strcpy(&tbuf[cmlen],arg);
+    cmlen+=strlen(arg)+1;
+    break;
+  case 7:  /* (string,string,string) */
+    typ=takearg(tedname, lno,ss,&arg,&tt);
+    if (typ!=2)
+      lper(tedname,lno,"String is expected");
+    strcpy(&tbuf[cmlen],arg);
+    cmlen+=strlen(arg)+1;
+    tt=skip_spc(tt);
+    if (*tt!=',') lper(tedname,lno,", is expected");
+    tt++;
+    typ=takearg(tedname, lno,tt,&arg,&tt);
+    if (typ!=2)
+      lper(tedname,lno,"String is expected");
+    strcpy(&tbuf[cmlen],arg);
+    cmlen+=strlen(arg)+1;
+    tt=skip_spc(tt);
+    if (*tt!=',') lper(tedname,lno,", is expected");
+    tt++;
+    typ=takearg(tedname, lno,tt,&arg,&tt);
+    if (typ!=2)
+      lper(tedname,lno,"String is expected");
+    strcpy(&tbuf[cmlen],arg);
+    cmlen+=strlen(arg)+1;
+    break;
+  case 8: /* int int */
+    typ=takearg(tedname, lno, ss,&arg,&tt);
+    if (typ!=1)
+      lper(tedname, lno, tedname,lno,"integer is expected");
+    ii=atoi(arg);
+    memcpy(&tbuf[cmlen],&ii,sizeof(int));
+		cmlen+=sizeof(int);
+		tt=skip_spc(tt);
+		if (*tt!=',') lper(tedname,lno,", is expected");
+		tt++;
+		ss=skip_spc(tt);
 		typ=takearg(tedname, lno, ss,&arg,&tt);
 		if (typ!=1)
 			lper(tedname, lno, tedname,lno,"integer is expected");
-		tbuf[cmlen++]=atoi(arg);
-		break;
-	case 2:
-		typ=takearg(tedname, lno, ss,&arg,&tt);
-		if (typ!=2)
-			lper(tedname,lno,"String is expected");
-		strcpy(&tbuf[cmlen],arg);
-		cmlen+=strlen(arg)+1;
-		break;
-	case 3:  /* (string,string) */
-		typ=takearg(tedname, lno,ss,&arg,&tt);
-		if (typ!=2)
-			lper(tedname,lno,"String is expected");
-		strcpy(&tbuf[cmlen],arg);
-		cmlen+=strlen(arg)+1;
-		tt=skip_spc(tt);
-		if (*tt!=',') lper(tedname,lno,", is expected");
-		tt++;
-		typ=takearg(tedname, lno,tt,&arg,&tt);
-		if (typ!=2)
-			lper(tedname,lno,"String is expected");
-		strcpy(&tbuf[cmlen],arg);
-		cmlen+=strlen(arg)+1;
-		break;
-	case 7:  /* (string,string,string) */
-		typ=takearg(tedname, lno,ss,&arg,&tt);
-		if (typ!=2)
-			lper(tedname,lno,"String is expected");
-		strcpy(&tbuf[cmlen],arg);
-		cmlen+=strlen(arg)+1;
-		tt=skip_spc(tt);
-		if (*tt!=',') lper(tedname,lno,", is expected");
-		tt++;
-		typ=takearg(tedname, lno,tt,&arg,&tt);
-		if (typ!=2)
-			lper(tedname,lno,"String is expected");
-		strcpy(&tbuf[cmlen],arg);
-		cmlen+=strlen(arg)+1;
-		tt=skip_spc(tt);
-		if (*tt!=',') lper(tedname,lno,", is expected");
-		tt++;
-		typ=takearg(tedname, lno,tt,&arg,&tt);
-		if (typ!=2)
-			lper(tedname,lno,"String is expected");
-		strcpy(&tbuf[cmlen],arg);
-		cmlen+=strlen(arg)+1;
-		break;
+    ii=atoi(arg);
+    memcpy(&tbuf[cmlen],&ii,sizeof(int));
+		cmlen+=sizeof(int);
 	}
 
 	ss=skip_spc(tt);
@@ -618,8 +705,10 @@ memcpy(*cmd,tbuf,cmlen);
 return 1;
 }
 
-static qcmp(KEYDEF *a, KEYDEF *b)
+static qcmp(const void *aa, const void *bb)
 {
+KEYDEF *a=(KEYDEF *)aa;
+KEYDEF *b=(KEYDEF *)bb;
 int i,mlen=Min(a->klen,b->klen);
 int res;
 for(i=0;i<mlen;i++)
@@ -628,9 +717,9 @@ if (i==mlen) return(a->klen - b->klen);
 return res;
 }
 
-static qcmpxk(KEYMAP *a, KEYMAP *b)
+static qcmpxk(const void *a, const void *b)
 {
-return(a->xkey - b->xkey);
+return(((KEYMAP *)a)->xkey - ((KEYMAP *)b)->xkey);
 }
 
 int file_exist(char *s)
@@ -647,7 +736,7 @@ int cdtedrc=1;
 FILE *init_tedrc_fname(char *ftedrc, int depth)
 {
 FILE *fp;
-char tt[80],uu[80];
+char tt[256],uu[256];
 #if	0
 printf(" %d %s\n",depth, ftedrc);
 #endif
@@ -662,6 +751,35 @@ if (fp=fopen(
 if (fp=fopen(strcat(strcat(strcpy(uu, TEDDIR),"/"),ftedrc),"r"))
 	return fp;
 if (!depth) {
+  FILE *tfp;
+  extern char *exec_name;
+  static char exepath[128];
+  int l;
+
+  if (exepath[0]) {
+    strcpy(tt,exepath);
+  } else {
+    sprintf(tt,"which %s",exec_name);
+    if ((tfp=popen(tt,"r"))==NULL)
+      p_err("cannot locate exec %s path",exec_name);
+    fgets(tt,sizeof(tt),tfp);
+    if ((l=strlen(tt))>0) {
+      int i=l-1;
+      while (i>=0 && tt[i]!='/') i--;
+      if (i>=0) tt[i]=0;
+      strcpy(exepath,tt);
+    }
+  }
+  if (fp=fopen(strcat(strcat(strcpy(uu, tt),"/"),ftedrc),"r"))
+    return fp;
+  if ((l=strlen(tt))>0) {
+    int i=l-1;
+    while (i>=0 && tt[i]!='/') i--;
+    if (i>=0) tt[i]=0;
+    strcat(tt,"/Ted");
+    if (fp=fopen(strcat(strcat(strcpy(uu, tt),"/"),ftedrc),"r"))
+      return fp;
+  }
 	if (cdtedrc)
 		p_err("Cannot find ./%s or %s or %s",ftedrc,tt,uu);
 	else
@@ -720,9 +838,9 @@ for(i=0;i<kdefno;i++) {
 return i;
 }
 
-static int kwcmp(KW *a, KW *b)
+static int kwcmp(const void *a, const void *b)
 {
-return strcmp(a->keyword,b->keyword);
+  return strcmp(((KW *)a)->keyword,((KW *)b)->keyword);
 }
 
 static int TedStart=1;
@@ -850,9 +968,11 @@ while (!feof(fp)) {
 		for(i=0;i<sizeof(sets)/sizeof(sets[0]);i++)
 		if (!strcmp(sets[i].sname,vv)) break;
 		if (i==sizeof(sets)/sizeof(sets[0]))
-			lper(tedname,lno,"Unknown set %s",vv);
-		*tt=0;
-		*(sets[i].saddr)=atoi(xx);
+			lper2(tedname,lno,"Unknown set %s",vv);
+		else {
+		  *tt=0;
+		  *(sets[i].saddr)=atoi(xx);
+		}
 	} else
 	if (!strcmp(tt,"lang") && !depth) {
 		extern char DirectCommentChars[];
@@ -944,6 +1064,12 @@ while (!feof(fp)) {
 		if (!(*vv)) lper(tedname,lno,"argument expected");
 		init_reg_exp(vv);
 	} else
+	if (!strcmp(tt,"fontlist")) {
+		ww=to_spc(vv);
+		xx=skip_spc(ww);
+		if (!(*vv)) lper(tedname,lno,"argument expected");
+		setFontList(vv);
+	} else
 	if (!strcmp(tt,"LoadKeyDefInto")) {
 		int no;
 		char mstr[16], *p;
@@ -999,18 +1125,22 @@ mode_str=mode_str_m[no];
 cmapno=no;
 }
 
+void set_kdef();
+
 K_LoadTedrc(char *ftedrc)
 {
 LoadTedrc(ftedrc,0,0);
-set_lang_type(fname);
 set_kbd_mode(0);
+set_lang_type(fname);
+set_kdef();
 }
 
 K_MergeTedrc(char *ftedrc)
 {
 LoadTedrc(ftedrc,1,0);
-set_lang_type(fname);
 set_kbd_mode(0);
+set_lang_type(fname);
+set_kdef();
 }
 
 void disp_kb_mode()
@@ -1081,12 +1211,12 @@ rec_mac=1;
 
 K_PlayMacro()
 {
-if (rec_mac) {
-	message(4,"You are still recording marcro");
-	return 0;
-}
-if (mac_buf && mac_cnt)
-	ExecCommand(mac_buf,mac_cnt);
+  if (rec_mac) {
+    message(4,"You are still recording marcro");
+    return 0;
+  }
+  if (mac_buf && mac_cnt)
+    ExecCommand(mac_buf,mac_cnt);
 }
 
 K_ExportMacro(char *fname)
@@ -1138,10 +1268,10 @@ fclose(fw);
 
 void ck_mac_size()
 {
-if (mac_ptr>=mac_buf+MAX_MACRO_BUF) {
-	message(4,"You forgot to stop macro recording");
-	rec_mac=0;
-}
+  if (mac_ptr>=mac_buf+MAX_MACRO_BUF) {
+    message(4,"You forgot to stop macro recording");
+    rec_mac=0;
+  }
 }
 
 cmdlen(char *cmd, int count)
@@ -1169,6 +1299,8 @@ for(i=0;i<count;i++) {
 		 ttt=cmd+strlen(cmd)+1;
 		 uuu=ttt+strlen(ttt)+1;
 		 cmd+=strlen(cmd)+1+strlen(ttt)+1+strlen(uuu)+1;
+	case 8:
+	   cmd+=sizeof(int)*2;
 		 break;
 	}
 }
@@ -1180,7 +1312,7 @@ int saveBufType;
 
 K_PutStrSave()
 {
-puts("should not enter here");
+  puts("should not enter here");
 }
 
 
@@ -1188,20 +1320,27 @@ int readonly=0;
 
 K_TogReadOnly()
 {
-readonly^=1;
-if (readonly)
-	message(4,"Read-only mode");
-else
-	message(4,"Read-Write mode");
+  readonly^=1;
+  if (readonly)
+    message(4,"Read-only mode");
+  else
+    message(4,"Read-Write mode");
 }
 
 static char *last_str;
 
+void X_loop()
+{
+while (input_handler) {
+	X_event();
+}
+}
+
 ExecCommand(char *cmd, int count)
 {
 	int i,usta,j;
 	char *ttt,*uuu;
-	u_char *pcmd;
+	char *pcmd;
 
 	pcmd=cmd;
 	last_str=0;
@@ -1219,6 +1358,7 @@ ExecCommand(char *cmd, int count)
 				ck_mac_size();
 			}
 			usta=func[funo].action();
+			X_loop();
 			break;
 		case 1:
 			if (rec_mac) {
@@ -1286,7 +1426,7 @@ ExecCommand(char *cmd, int count)
 		case 7:
 			ttt=cmd+strlen(cmd)+1;
 			uuu=ttt+strlen(ttt)+1;
-			usta=func[funo].action(cmd,ttt,uuu);
+      usta=func[funo].action(cmd,ttt,uuu);
 			if (rec_mac) {
 				*(mac_ptr++)=funo;
 				strcpy(mac_ptr, cmd);
@@ -1299,6 +1439,19 @@ ExecCommand(char *cmd, int count)
 			}
 			cmd+=strlen(cmd)+1+strlen(ttt)+1+strlen(uuu)+1;
 			break;
+		case 8:
+		{
+		  int aa,bb;
+		  memcpy(&aa,cmd,sizeof(int));
+		  memcpy(&bb,cmd+4,sizeof(int));
+      usta=func[funo].action(aa,bb);
+      cmd+=sizeof(int)*2;
+    }
+		  break;
+		}
+    if (xgets_enter==-1) {
+		  xgets_enter=0;
+		  break;
 		}
 		if (usta) continue;
 		disp_fname();
